/****(c) 2020-2021. A. Gerbessiotis. See ../KAagcopy.txt ****/
/************************************************************/
/*
#include <math.h>
#include <time.h>  if you need srand_r(time(0))
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "aggssrt.h"

#define ALL_FUNCTIONS
#undef  ALL_FUNCTIONS
#undef  FEW_FUNCTIONS
#define FEW_FUNCTIONS
#define NALGS 35   /* Make sure # entries in gfunc/doalg is  NALGS or CRASH!*/

#define ELMNTS 8
#define MODULO 100000
#define MODULO_UINT ((unsigned int) 0x7FFFFFFF)
#define AVG_TINIT   AVG_PERRUN = AVG_ELAPSED =(double)0;
#define AVG_TSTART  AVG_PERRUN = (double) 0.0 ; avg_tstart();
#define AVG_TEND    avg_tend(); AVG_PERRUN = (avg_treturn() ); AVG_ELAPSED +=(AVG_PERRUN);
#define AVG_PERRUN  per_run
#define AVG_ELAPSED elapsed

static struct timeval avgtstart, avgtend;
static struct timezone avgtstruct;

void   avg_tstart  (void);
void   avg_tend    (void);
double avg_treturn (void);
void   printseq    (gdata *inp,int n,char *s,int debug);
void   dataseq     (gdata *d, int n, int type);
void   datacheck   (gdata *res, gdata *ori , int n);
void   seqdatacheck(gdata *ori , int n);

void   (*GTYPESORT)(void *,int,int,int (*) (const void *,const void *));


void avg_tstart(void) {
  gettimeofday(&avgtstart, &avgtstruct);
}

void avg_tend(void) {
  gettimeofday(&avgtend,&avgtstruct);
}

double avg_treturn(void) {
  double t1,t2;
  t1=(double)avgtstart.tv_sec + (double)avgtstart.tv_usec*(0.000001);
  t2=(double)  avgtend.tv_sec + (double)  avgtend.tv_usec*(0.000001);
  return(t2-t1);
}

int g1compare(const void *x,const void *y)
{
	return(int)(* ((gdata*)x)-*((gdata*)y));
}

int main(int argc,char *argv[]) {
  int j,k,n,runs, type, debug;
  gdata *ptr,*bptr, *sptr;
  double etime;

  void (*gfunc[NALGS])() = {  
    qsort,    g0gnu1,    g6qkrw1,  g6qkrw2, g6qkrw3,  
    g6qkrw4,  g6qkrw5,   g7qkrw1,  g7qkrw2, g7qkrw3, 
    g7qknd0,  g7qknd1,   g7qknd2,  g7qknd3, g7qknd4, 
    g7qknd5,  g7qknd6,   g7qknd7,  g6qkrd1, g6qkrd2,   
    g3mrgiu,  g3mrgio,   g5hpx1,   g5hpx2,  g5hpx3,    
    g1inb1,   g1ins1,    g1ins2,   qsort,   g2bub1,   
    g2bub2,   g2bub3,    g2oets,   g6carh1, g6carh2
   };

  char *labels[NALGS]   = { 
    "qsort",    "g0gnu1",   "g6qkrw1",  "g6qkrw2",  "g6qkrw3",  
    "g6qkrw4",  "g6qkrw5",  "g7qkrw1",  "g7qkrw2",  "g7qkrw3",  
    "g7qknd0",  "g7qknd1",  "g7qknd2",  "g7qknd3",  "g7qknd4",
    "g7qknd5",  "g7qknd6",  "g7qknd7",  "g6qkrd1",  "g6qkrd2",  
    "g3mrgiu",  "g3mrgio",  "g5hpx1",   "g5hpx2",   "g5hpx3",   
    "g1inb1",   "g1ins1",   "g1ins2",   "qsort",    "g2bub1",   
    "g2bub2",   "g2bub3",   "g2oets",   "g6carh1",  "g6carh2"
   };

  int doalg[NALGS] = { 
/*
    1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
    1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
    1,1,1,1,1

    1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
    1,1,1,1,1,  1,1,1,1,1,  0,0,0,1,0,
    0,0,0,1,1

    1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
    1,1,1,1,1,  1,1,1,1,1,  0,0,0,1,0,
    0,0,0,1,1

*/
    1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
    1,1,1,1,1,  1,1,1,1,1,  1,1,1,1,1,
    1,1,1,1,1
   };

   ptr=bptr=sptr=NULL;

   n=1000;   runs=1;
   if (( argc <= 1 ) || (argc > 5)) {
             printf("Usage: %s n type runs debug\n",argv[0]);
             printf("Type : 1=[0..RAND_MAX==?2**31-1], 2=[0..n],");
             printf(" 3= INC, 4= 17  else= DEC\n");
             printf("Quick:  sgMRG=%d sgQUI=%d SPLITTER=%d\n",
                     SGMRG, SGQUI,  SPLITTER); 
             printf("Split: 0:rand 1:left, 2:middle, 3:Med3(some)\n");
             return(1);
   } 
     else {
             n    = atoi(argv[1]);  /*  size  */
             type = atoi(argv[2]);  /*  type  */
             runs = atoi(argv[3]);  /*  runs  */
             debug= atoi(argv[4]);  /*  debug */
             printf("%d : %d %d %d %d (time in ms)\n",argc,n,type,runs,debug);
   }
   etime=0.0;
   sptr= (gdata *) malloc(n*sizeof(gdata));
   bptr= (gdata *) malloc(n*sizeof(gdata));
   ptr = (gdata *) malloc(n*sizeof(gdata));
   memset( ptr,0,n*sizeof(gdata));
   memset(bptr,0,n*sizeof(gdata));
   memset(sptr,0,n*sizeof(gdata));

   dataseq(bptr,n,type);
   printseq(bptr,n,"InQ",debug);
   memcpy(ptr,bptr,n*sizeof(gdata));

   qsort(ptr,n,sizeof(gdata),g1compare);
   printseq( ptr,n,"OtQ",debug);
   memcpy(sptr,ptr,n*sizeof(gdata));

   for(k=0;k<NALGS;k++) {
      if (doalg[k]>=1) {
         GTYPESORT=gfunc[k];
         etime=0.0;

         for(j=0;j<runs;j++) {
            memcpy(ptr,bptr,n*sizeof(gdata));
            avg_tstart();
               GTYPESORT(ptr,n,sizeof(gdata),g1compare);
            avg_tend();
            etime += (avg_treturn());
         }
/*       memcmp 3 lines below does not tell you which keys are out of sequence
*/
         datacheck(ptr,sptr,n);

      if (memcmp(ptr,sptr,n*sizeof(gdata))  != 0){
         printf("NO:%2d:%2d:%1d:%10d:%1d:%2d:%1d:",SGMRG  ,SGQUI ,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.5f : %12.5f\n",labels[k],1000*etime,1000*etime/runs);
      }
      else {
         printf("%2d:%2d:%1d:%10d:%1d:%2d:%1d:",SGMRG  ,SGQUI  ,SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.5f : %12.5f\n",labels[k],1000*etime,1000*etime/runs);
         fflush(stdout);
      }
      }
   }

   free((void *) ptr);
   free((void *) bptr);
   free((void *) sptr);
   return(0);
}

void printseq(gdata *inp, int n, char *s, int debug) 
{
 register int i;
 if (1==debug)  { 
 fprintf(stdout,"[%4s]: ",s);
 fflush(stdout);
 for(i=0;i<ELMNTS;i++){
   fprintf(stdout,"%10d,", (gdata) inp[i]);
 }
   fprintf(stdout,"...\n          ");
 for(i=n-ELMNTS;i<n;i++){
   fprintf(stdout,"%10d,", (gdata) inp[i]);
 }
 fprintf(stdout,"\n");
 fflush(stdout);
 }
}

void dataseq  (gdata *d, int n, int type)
{
 register int i;
 unsigned int seed =rand() ; /* 21+1001*11 */
/*
 srand(time(0));
*/
 seed = rand(); /* same value; uncomment line above otherwise */
    if (type == 1) {
        for (i=0;i<n;i++) {
           d[i]= (gdata ) (rand_r(&seed));
        }
     } 
     else 
     if (type == 2) {
         for (i=0;i<n;i++) {
            d[i]= (gdata )  ((n)*(rand_r(&seed)/((double)RAND_MAX)));
         }
     }
     else
     if (type == 3) {
         for (i=0;i<n;i++) {
            d[i]= (gdata )  (i+1);
         }
     }
     else
     if (type == 4) {
         for (i=0;i<n;i++) {
            d[i]= (gdata )  (17);
         }
     }
     else {
       for (i=0;i<n;i++) {
         d[i]= (gdata )  (n-i);
       }
     }
}



void seqdatacheck (gdata *ori, int n)
{
  register int i;
   for(i=0;i<n-1;i++) {
     if (ori[i] > ori[i+1]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,ori[i],ori[i+1]);
      fflush(stdout); 
      return;
     }
   }
   fprintf(stdout,"OK:"); fflush(stdout);
}

void datacheck (gdata *res, gdata *ori , int n)
{
  register int i;
   for(i=0;i<n;i++) {
     if (res[i] != ori[i]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,res[i],ori[i]);
      fflush(stdout);
      return;
     }
   }
   fprintf(stdout,"OK:");fflush(stdout);
}

/* RETIRED
    srandom(21+1001*11);
    if (type == 1) {
        for (i=0;i<n;i++) {
           d[i]= (gdata ) (random() & MODULO_UINT);
        }
     } 
     else 
     if (type == 2) {
         for (i=0;i<n;i++) {
            d[i]= (gdata )  ((n)*(random()/((double)INT_MAX)));
         }
     }
*/
