/****(c) 2020-2021. A. Gerbessiotis. See ../KAagcopy.txt ****/
/************************************************************/
/*
#include <math.h>
#include <time.h>  if you need srand_r(time(0))
*/
#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include "agissrt.h"

#define ALL_FUNCTIONS
#undef  ALL_FUNCTIONS
#undef  FEW_FUNCTIONS
#define FEW_FUNCTIONS
#define NALGS 17   /* Make sure #entries in gfunc/doalg is NALGS or CRASH!!!! */

#define ELMNTS 8
#define MODULO 100000
#define MODULO_UINT ((unsigned int) 0x7FFFFFFF)
#define AVG_TINIT   AVG_PERRUN = AVG_ELAPSED =(double)0;
#define AVG_TSTART  AVG_PERRUN = (double) 0.0 ; avg_tstart();
#define AVG_TEND    avg_tend(); AVG_PERRUN = (avg_treturn() ); AVG_ELAPSED +=(AVG_PERRUN);
#define AVG_PERRUN  per_run
#define AVG_ELAPSED elapsed

static struct timeval avgtstart, avgtend;
static struct timezone avgtstruct;

void   avg_tstart  (void);
void   avg_tend    (void);
double avg_treturn (void);
void   iprintseq    (idata *inp,int n,char *s,int debug);
void   idataseq     (idata *d, int n, int type);
void   idatacheck   (idata *res, idata *ori , int n);
void   iseqdatacheck(idata *ori , int n);

void   (*ITYPESORT)(idata *,int);


void avg_tstart(void) {
  gettimeofday(&avgtstart, &avgtstruct);
}

void avg_tend(void) {
  gettimeofday(&avgtend,&avgtstruct);
}

double avg_treturn(void) {
  double t1,t2;
  t1=(double)avgtstart.tv_sec + (double)avgtstart.tv_usec*(0.000001);
  t2=(double)  avgtend.tv_sec + (double)  avgtend.tv_usec*(0.000001);
  return(t2-t1);
}

int g1compare(const void *x,const void *y)
{
	return(int)(* ((idata*)x)-*((idata*)y));
}

int main(int argc,char *argv[]) {
  int j,k,n,runs, type, debug;
  idata *ptr,*bptr, *sptr;
  double etime;

  void (*ifunc[NALGS])() = {  
   i6clrs,   i6qkrd1, i6qkrf1, i6qkrw1, i9rdx4u, 
   i7qknd1,  i7qknd2, i9rdx4u, i9rdx4o, i4mrgi1, 
   i5hpx1,  i5hpx3,   i1ins1,   i1ins2, i1ins3,  
   i1ins4,  i1ins5,
   };

  char *labels[NALGS]   = { 
   "i6clrs",  "i6qkrd1", "i6qkrf1", "i6qkrw1", "i9rdx4u",
   "i7qknd1", "i7qknd2", "i9rdx4u", "i9rdx4o", "i4mrgi1", 
   "i5hpx1",  "i5hpx3",  "i1ins1",  "i1ins2",  "i1ins3",  
   "i1ins4",  "i1ins5",
   };

  int doalg[NALGS] = { 
   1,1,1,1,1,   1,1,1,1,1,   1,1,1,1,1, 1,1
/*
   0,0,0,0,0,   1,1,1,1,1,   1,1,1,1,1, 0,0
   1,1,1,1,1,   1,1,1,1,0,   0,0,0,0,0, 0,0
*/
   };

   ptr=bptr=sptr=NULL;

   n=1000;   runs=1;
   if (( argc <= 1 ) || (argc > 5)) {
             printf("Usage: %s n type runs debug\n",argv[0]);
             printf("Type : 1=[0..RAND_MAX==?2**31-1], 2=[0..n],");
             printf(" 3= INC, 4= 17  else= DEC\n");
             printf("SDMRG=%d SIMRG=%d SDQUI=%d SIQUI=%d SPLIT(l,m,rand)==%d\n",
                     SDMRG,SIMRG,SDQUI,SIQUI,SPLITTER);
             return(1);
   } 
     else {
             n    = atoi(argv[1]);  /*  size  */
             type = atoi(argv[2]);  /*  type  */
             runs = atoi(argv[3]);  /*  runs  */
             debug= atoi(argv[4]);  /*  debug */
             printf("%d : %d %d %d %d\n",argc,n,type,runs,debug);
   }
   etime=0.0;
   sptr= (idata *) malloc(n*sizeof(idata));
   bptr= (idata *) malloc(n*sizeof(idata));
   ptr = (idata *) malloc(n*sizeof(idata));
   memset( ptr,0,n*sizeof(idata));
   memset(bptr,0,n*sizeof(idata));
   memset(sptr,0,n*sizeof(idata));

   idataseq(bptr,n,type);
   iprintseq(bptr,n,"InQ",debug);
   memcpy(ptr,bptr,n*sizeof(idata));

   qsort(ptr,n,sizeof(idata),g1compare);
   iprintseq( ptr,n,"OtQ",debug);
   memcpy(sptr,ptr,n*sizeof(idata));


  etime=0.0;
  for(j=0;j<runs;j++){
       memcpy(ptr,bptr,n*sizeof(idata));
       avg_tstart();
          qsort(ptr,n,sizeof(idata),g1compare);
       avg_tend();
       etime += (avg_treturn());
  }
  idatacheck(ptr,sptr,n);

  if (memcmp(ptr,sptr,n*sizeof(idata))  != 0){
     printf("NO:%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
            SIMRG,SIQUI,SPLITTER,n,type,runs,debug);
     printf(   "%11s : %12.8f : %12.3f\n","qsort",etime,etime/runs);
  }
  else {
     printf("%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
            SIMRG,SIQUI,SPLITTER,n,type,runs,debug);
     printf(   "%11s : %12.8f : %12.3f\n","qsort",etime,etime/runs);
     fflush(stdout);
  }
  etime=0.0;
  for(j=0;j<runs;j++){
       memcpy(ptr,bptr,n*sizeof(idata));
       avg_tstart();
          i0gnu(ptr,n,sizeof(idata),g1compare);
       avg_tend();
       etime += (avg_treturn());
  }
  idatacheck(ptr,sptr,n);

  if (memcmp(ptr,sptr,n*sizeof(idata))  != 0){
     printf("NO:%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
              SIMRG, SIQUI,   SPLITTER,n,type,runs,debug);
     printf(   "%11s : %12.8f : %12.3f\n","i0gnu",etime,etime/runs);
  }
  else {
     printf("%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
              SIMRG, SIQUI,   SPLITTER,n,type,runs,debug);
     printf(   "%11s : %12.8f : %12.3f\n","i0gnu",etime,etime/runs);
     fflush(stdout);
  }

   for(k=0;k<NALGS;k++) {
      if (doalg[k]>=1) {
         ITYPESORT=ifunc[k];
         etime=0.0;

         for(j=0;j<runs;j++) {
            memcpy(ptr,bptr,n*sizeof(idata));
            avg_tstart();
               ITYPESORT(ptr,n);
            avg_tend();
            etime += (avg_treturn());
         }
/*       memcmp 3 lines below does not tell you which keys are out of sequence
*/
         idatacheck(ptr,sptr,n);

      if (memcmp(ptr,sptr,n*sizeof(idata))  != 0){
         printf("NO:%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
                SIMRG, SIQUI,  SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n",labels[k],etime,etime/runs);
      }
      else {
         printf("%2d:%2d:%1d:%10d:%1d:%2d:%1d:",
                SIMRG, SIQUI,   SPLITTER,n,type,runs,debug);
         printf(   "%11s : %12.8f : %12.3f\n",labels[k],etime,etime/runs);
         fflush(stdout);
      }
      }
   }

   free((void *) ptr);
   free((void *) bptr);
   free((void *) sptr);
   return(0);
}

void iprintseq(idata *inp, int n, char *s, int debug) 
{
 register int i;
 if (1==debug)  { 
 fprintf(stdout,"[%4s]: ",s);
 fflush(stdout);
 for(i=0;i<ELMNTS;i++){
   fprintf(stdout,"%10d,", (idata) inp[i]);
 }
   fprintf(stdout,"...\n          ");
 for(i=n-ELMNTS;i<n;i++){
   fprintf(stdout,"%10d,", (idata) inp[i]);
 }
 fprintf(stdout,"\n");
 fflush(stdout);
 }
}

void idataseq  (idata *d, int n, int type)
{
 register int i;
 unsigned int seed = rand(); /* 21+1001*11 */
 /*
  srand(time(0));
  */
 seed=rand(); /* 21+1001*11 */
    if (type == 1) {
        for (i=0;i<n;i++) {
           d[i]= (idata ) (rand_r(&seed));
        }
     } 
     else 
     if (type == 2) {
         for (i=0;i<n;i++) {
            d[i]= (idata )  ((n)*(rand_r(&seed)/((double)RAND_MAX)));
         }
     }
     else
     if (type == 3) {
         for (i=0;i<n;i++) {
            d[i]= (idata )  (i+1);
         }
     }
     else
     if (type == 4) {
         for (i=0;i<n;i++) {
            d[i]= (idata )  (17);
         }
     }
     else {
       for (i=0;i<n;i++) {
         d[i]= (idata )  (n-i);
       }
     }
}



void iseqdatacheck (idata *ori, int n)
{
  register int i;
   for(i=0;i<n-1;i++) {
     if (ori[i] > ori[i+1]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,ori[i],ori[i+1]);
      fflush(stdout); 
      return;
     }
   }
   fprintf(stdout,"OK:"); fflush(stdout);
}

void idatacheck (idata *res, idata *ori , int n)
{
  register int i;
   for(i=0;i<n;i++) {
     if (res[i] != ori[i]) {
      fprintf(stdout,"\nError at %d : %d noteq %d\n",i,res[i],ori[i]);
      fflush(stdout);
      return;
     }
   }
   fprintf(stdout,"OK:");fflush(stdout);
}

