/****Copyright (c) 1997-2004 All rights reserved
 ****A. V. Gerbessiotis and C. J. Siniolakis
 ****Permission to use, copy, modify, and distribute this software,
 ****and to incorporate it, in whole or in part, into 
 ****other software,is hereby granted without fee, provided that
 ****  (1) the above copyright notice and this permission 
 ****      notice appear in all copies of the source code, and 
 ****      the above copyright notice appear in clearly visible 
 ****      form on all supporting documentation and distribution 
 ****      media, and
 ****     (2) any redistribution of the software, in original
 ****         or modified form, be without fee and subject to
 ****         these same conditions.
 ****   No guarantee, whatsoever, is offered that the code works as
 ****   advertised or is absent of any, even damaging,
 ****   side-effects. Use this code at your own personal risk.
 ****/
#include "ssrt.h"
/* This is g68wbh; replicated for completeness */


void A268wbh(void *akeys, int   n,int size,int (*compare)(const void*,const void*), int *L,int *R)
{
  register int  M,c,I,J,l,r,mL,mR;
  register char *tmp,*pi,*pj,*pl,*pr,*split;
  char *keys,*DES,*SRC;

  if ((*R-*L+1) <= 1) return ;

  keys = (char *) akeys;
  tmp  = (char *) malloc(size*sizeof(char));
  split= (char *) malloc(size*sizeof(char));
#ifdef SANITY_CHECK
  check_if_null(split,"g268wbh","split");
  check_if_null(tmp  ,"g268wbh",  "tmp  ");
#endif
   mL=*L;mR=*R;
  while (mL < mR ) { 
     l=mL;r=mR;
     pl=&keys[l*size];
     pr=&keys[r*size];
#if SPLITTER == MIDDLE_KEY
    MEMCPY(split,&keys[((l+r)>>1)*size],size);
#elif SPLITTER == FIRST_KEY
    MEMCPY(split,&keys[l*size],size);
#elif SPLITTER == RANDOM_KEY
    MEMCPY(split,&keys[(l+GRANDOM((r-l+1)))*size],size);
#endif

     while (l <= r) {
         while (compare(pl,split)<0) {l++;pl +=size;}
         while (compare(split,pr)<0) {r--;pr -=size;}
         if (l <= r) { /* is it ok if < ? */
/*
	     MEMCPY(tmp,pl,size);
	     MEMCPY(pl,pr,size);
	     MEMCPY(pr,tmp,size);
    	     l++;r--;pl += size; pr -= size;
*/
               M=size;
               do {
                c=*pl; *pl++ = *pr; *pr++=c;
               } while (--M);
               l++; pr -=size;
               r--; pr -=size;

         }
     }
  
   if ((r-mL+1) < (mR-l+1)) { /* Left side smaller */
      if ((r-mL+1) < THRES16) {
                 if ((r-mL+1) >1) {
                  pi=&keys[mL*size];
                  for(J=1;J<(r-mL+1);J++){
                    MEMCPY(split,(pi+J*size),size);
                    I=J-1;pj=pi+I*size;
                    while ((I>=0) && (compare(pj,split)>0) ) {
                     MEMCPY((pj+size),(pj),size);
                     I--;pj-=size;
                    }
                    MEMCPY((pj+size),split,size);
                  }
                 }
          mL=l;
      }
      else {
         *R=r; *L=mL; /* mL...r */
         A268wbh(keys,(r-mL+1),size,compare,L,R);
         mL=l;  /* OR r l ..mR */
      }
   } /*left */
  else {               /* Right side smaller */
      if ((mR-l+1) < THRES16) {
                          if ((mR-l+1) >1) {
               pi=&keys[(l)*size];
               for(J=1;J<(mR-l+1);J++){
                 MEMCPY(split,(pi+J*size),size);
                 I=J-1;pj=pi+I*size;
                 while ((I>=0) && (compare(pj,split)>0) ) {
                  MEMCPY((pj+size),(pj),size);
                  I--;pj-=size;
                 }
                 MEMCPY((pj+size),split,size);
            }
           }
       mR=r;
      }
      else {
        *L=l; *R=mR;  /* l...mR */
        A268wbh(keys,mR-l+1,size,compare,L,R);
        mR=r;  /* mL ...r */
      }
      
  } /*right*/
 }


#ifdef ZERO
  if (r > 0) {
      if ((r+1)<THRES16) {
            if ((r+1) >1) {
               pi=keys;
               for(J=1;J<(r+1);J++){
                 MEMCPY(split,(pi+J*size),size);
                 I=J-1;pj=pi+I*size;
                 while ((I>=0) && (compare(pj,split)>0) ) {
                  MEMCPY((pj+size),(pj),size);
                  I--;pj-=size;
                 }
                 MEMCPY((pj+size),split,size);
            }
           }
      }
      else A268wbh(keys,r+1,size,compare);
  }
  if (l < n) {
      if ((n - l)<THRES16) {
            if ((n-l) >1) {
               pi=&keys[l*size];
               for(J=1;J<(n-l);J++){
                 MEMCPY(split,(pi+J*size),size);
                 I=J-1;pj=pi+I*size;
                 while ((I>=0) && (compare(pj,split)>0) ) {
                  MEMCPY((pj+size),(pj),size);
                  I--;pj-=size;
                 }
                 MEMCPY((pj+size),split,size);
            }
           }
      }
      else A268wbh(&keys[l*size],n-l,size,compare);
  }
#endif
 free((void *)split); free((void *)tmp);
}
           
void g268wbh(void *keys, int n, int size, int (*compare) (const void*,const void *))
{
  int myL,myR;
  if (n <= 1) return ;
#if SPLITTER == RANDOM_KEY
 GSRANDOM();
#endif
  myL=0;myR=n-1;
  A268wbh(keys,n,size,compare,&myL,&myR);
}



/************************
 * Note : code differs from CLR Hoare code in the following respect
 *  say input 10 9 8 7 5 and splitter is middle key 8
 *  at termination we have say l=1 r=5 in the beginning
 *            6 7  8 9 10
 *              r    l
 *  lest partition is 6 7 and right partion is ? 9 10 or  8 9 10?
 *  if all keys are equal
 *            8 8  8  8 8 
 *              r     l   at termination
 *  In the g67.c code which is similar (but original and possibly primitive)
 *  left set is L ... r of size r-L+1   base left set keys[L*size]
 *  righ set is l ... R of size R-l+1 correctly identifies keys[l*size]
 *  if g68 and also by extension in g268 we used (pending before potential changes)
 *  left set is size r+1 and right set of size n-l which does not make much sense.
 */
